//
// ***************************************************************************
// *                                                                         * 
// *  Copyright Compaq Computer Corporation 2000.     All rights reserved    *
// *                                                                         *  
// ***************************************************************************
//
//
//   This is an example program to illustrate the use of the iterative 
//   solver ditsol_pcg from a C application program. The program generates 
//   the matrix and the preconditioner, calls the solver and prints the 
//   maximum error in the solution. The right hand side of the problem is 
//   generated assuming a known solution. The problem used is identical to 
//   the one in the example section of the chapter on iterative solvers
//   in the CXML Reference Guide.
//
//   This program illustrates the following:
//        - routine naming conventions
//        - Differences in endexing arrays:
//              C default: x[n] -> 0 to (n-1)
//              FORTRAN default: x(n) -> 1 to n
//        - how to use two dimensional arrays in C to interface with a
//          Fortran library routine
//        - how to use the matrix-free formulation from a C program
// 
//    For more detailed explanation of the routines used, please
//    check the CXML Reference Manual.
//
//    Note: the code used in this example works on all Compaq platforms.
//    Conditional compilation is used to select the statements 
//    appropriate to each operating system.
//
//    All output from this program is sent to the screen. A sample 
//    output is:
//
//   method used : cg with spd split preconditioning
//   order of system =       100
//   stopping criterion used =     1
//   maximum iterations allowed =         100
//   tolerance for convergence =   0.10000000E-05
//      iteration =          0    stopping test = 0.69282032E+01
//      iteration =          1    stopping test = 0.19194193E+01
//      iteration =          2    stopping test = 0.12100937E+01
//      iteration =          3    stopping test = 0.52439623E+00
//      iteration =          4    stopping test = 0.84029860E-01
//      iteration =          5    stopping test = 0.20539881E-01
//      iteration =          6    stopping test = 0.34309306E-02
//      iteration =          7    stopping test = 0.47063334E-03
//      iteration =          8    stopping test = 0.16605002E-03
//      iteration =          9    stopping test = 0.45072557E-04
//      iteration =         10    stopping test = 0.72087304E-05
//      iteration =         11    stopping test = 0.88047573E-06
//   solution obtained after          11  iterations
//   normal exit from solver
//   final value of stopping test = 0.88047573E-06
//
//maximum error in the solution: 5.6260082260e-08     
//
  
#include <iostream.h>
#include <iomanip.h>
#include <stdlib.h>
#include <new.h>


//
//   Add trailing underscores to Fortran routines on unix and linux
//

#if defined(linux)
#  define ditsol_defaults ditsol_defaults__
#  define dcreate_ilu_sdia dcreate_ilu_sdia__
#  define ditsol_pcg ditsol_pcg__
#  define dmatvec_sdia dmatvec_sdia__
#  define dapply_ilu_sdia dapply_ilu_sdia__
#elif defined(__unix__)
#  define ditsol_defaults ditsol_defaults_
#  define dcreate_ilu_sdia dcreate_ilu_sdia_
#  define ditsol_pcg ditsol_pcg_
#  define dmatvec_sdia dmatvec_sdia_
#  define dapply_ilu_sdia dapply_ilu_sdia_
#endif

inline double ABS(double x) 
{
    return(((x) < 0) ? -(x) : (x));
}    

inline double MAX(double x, double y) 
{
    return(((x) < (y)) ? (y) : (x));    
}

//  extern void (*set_new_handler(void (*memory_err)()))();
void memory_err() 
{
    cout << "memory allocation error\n";
    exit(1); // quit program
}

extern void pcondl1(int &, int [], double [], 
                    double [], int [], 
                    double [], int [], 
                    double [], double [], double [], int &); 
extern void  matvec1(int &, int [], double [], 
                     double [], int [], 
                     double [], double [], double [], int &);
extern void  genmat1(int, int, int, 
                     double [], int [],
                     int, int);

//
//   Declare the Fortran routines
//
              
extern "C" 
{
void ditsol_defaults(int [], double []);
void dcreate_ilu_sdia(double [], int [], 
                      int &, int &, 
                      double [], int [],
                      int &);
void ditsol_pcg(void (*)(int &, int [], double [], 
                         double [], int [], 
                         double [], double [], double [], 
                         int &), 
                 void (*)(int &, int [], double [], 
                          double [], int [], 
                          double [], int [], 
                          double [], double [], double [], int &), 
                 double &, double &,
                 double [], int [],
                 double [], double [], int &,
                 double [], int [], 
                 double &, int &, 
                 int [], double [],  
                 int &, double [], 
                 int &);
void dmatvec_sdia(int &, double [], int [],
                  int &, int &, double [], double [], double [], 
                  int &, int []);
void dapply_ilu_sdia(int &, 
                     double [], int [], 
                     int [], int [], 
                     double [], double [], int &);
}

//
//     illustrating the use of the iterative solver:
//     preconditioned conjugate gradient method, with incomplete
//     cholesky preconditioning. the matrix is stored using the
//     symmetric diagonal format
//

void main()
{
    double *a_sdia;
    double *a_ilu;
    double *rwork1;
    double *rhs;
    double *x;
    double *xo;
 
    double rparam[50];

    double dum, max1, tmp1;

    int *index_sdia;
    int *index_ilu;

    int iparam[50]; 
    
    int nxny, length, ndim, nzeros;
    int i, j, idum, ierror;
    int job; 

// set up exception handler

    set_new_handler(memory_err);
    
// define the problem size

    const int nx = 10;
    const int ny = 10;

    nxny = nx * ny;
    ndim = nxny;
    nzeros = 3;

// allocate memory for the 2-dimensional arrays

    a_sdia = new double [nzeros*ndim];
    a_ilu = new double [nzeros*ndim];

// allocate memory for the 1-dimensional arrays

    rwork1 = new double [4*nxny];
    rhs = new double [nxny];
    x = new double [nxny];
    xo = new double [nxny];
    
    index_sdia = new int [nzeros];
    index_ilu = new int [nzeros];
    
// set the parameters (integer and real) 

    ditsol_defaults(iparam, rparam); 

    iparam[2] = 0;
    iparam[3] = 4 * nxny;

// direct all output to the screen 

    iparam[4] = 6;
    iparam[5] = 3;
    iparam[6] = 4;

// generate the matrix 

    genmat1(nx, ny, nxny, 
            a_sdia, index_sdia, 
            ndim, nzeros); 

    iparam[30] = nzeros;
    iparam[31] = ndim;

// generate xo, the true solution 

    for (i=0; i<nxny; i++)
        xo[i] = 1.0; 

// obtain the right hand side 

    job = 0; 
    matvec1(job, iparam, rparam, 
            a_sdia, index_sdia, 
            &dum, xo, rhs, nxny); 
    
// obtain initial guess (all zeros) 

    for (i=0; i<nxny; i++)
        x[i] = 0.0; 
    
// generate the preconditioner 

    dcreate_ilu_sdia(a_sdia, index_sdia, 
                     ndim, nzeros, 
                     a_ilu, index_ilu, 
                     nxny);

// call the solver 

    ditsol_pcg(matvec1, pcondl1, dum, dum,
               a_sdia, index_sdia,
               x, rhs, nxny,
               a_ilu, index_ilu, 
               dum, idum,
               iparam, rparam, 
               idum, rwork1, 
               ierror); 

    if (ierror != 0) 
        cout << "ditsol_pcg returned with error flag: " << ierror 
            << endl;

// find the maximum absolute error in the solution 

    max1 = ABS((x[0]-xo[0]));

    for (i=1; i<nxny; i++)
        {
            tmp1 = ABS((x[i]-xo[i]));
            max1 = MAX((max1),(tmp1));      
        }
// print the maximum absolute error 

    cout << "maximum error in the solution: " << max1 << endl;

// deallocate the memory

    delete a_sdia;
    delete a_ilu;
    delete rwork1;
    delete xo;
    delete x;
    delete rhs;
    
    delete index_sdia;
    delete index_ilu;
    
    
}  // end of main() 



//
//    generate the matrix for the problem described in the chapter on 
//    iterative solvers in the CXML Reference Guide
//

void genmat1(int nx, int ny, int nxny, 
             double a[], int index[], 
             int ndim, int nzeros)
{
    int  i, j, k; 
    
    for (j=0; j<nxny; j++)
        for (i=1; i<nzeros; i++)
            a[i*ndim+j] = 0.0; 
    
    for (j=0; j<nxny; j++)
        a[0*ndim+j] = 4.0; 

    for (j=0; j<ny; j++)
        for (i=1; i<nx; i++)
            {
                k = j * nx + i; 
                a[2*ndim+k] = -1.0; 
            }

    for (j=1; j<ny; j++)
        for (i=0; i<nx; i++)
            {
                k = j * nx + i; 
                a[1*ndim+k] = -1.0; 
            }
    
    index[0] = 0;
    index[2] = -1;
    index[1] = -nx;

}  // end of genmat1() 



//
//    provide the matrix-vector routine using the standard parameter list
//    as described in the CXML Reference Guide
//

void matvec1(int &job, int iparam[], double rparam[], 
             double a[], int ia[], 
             double w[], double x[], double y[], int &n)
{
    int  nzeros, ndim; 
    double dum; 
 
    nzeros = iparam[30];
    ndim = iparam[31];

    dmatvec_sdia(job, a, ia, ndim, nzeros, &dum, x, y, n, iparam); 

}  // end of matvec1()



//
//    provide the left preconditioning routine using the standard parameter
//    list as described in the CXML Refernce Guide
//

void pcondl1(int &job, int iparam[], double rparam[], 
             double ql[], int iql[], 
             double a[], int ia[], 
             double w[], double x[], double y[], int &n)
{
    int  nzeros, ndim, job1; 
    double *tmp;
    
// ilu preconditioning 

    nzeros = iparam[30];
    ndim = iparam[31];

// allocate temporary storage

    tmp = new double [n];

    job1 = 0; 
    dapply_ilu_sdia(job1, ql, iql, &ndim, &nzeros, x, tmp, n);
    job1 = 1; 
    dapply_ilu_sdia(job1, ql, iql, &ndim, &nzeros, tmp, y, n);

// deallocate temporary storage 

    delete tmp;
    
}  // end of pcondl1() 




