/*
 *****************************************************************************
 *                                                                           *
 *  Copyright Compaq Computer Corporation 1999.         All rights reserved. *
 *                                                                           *
 *****************************************************************************
*/
/*
    This is an example program to illustrate the use of the lapack
    linear solver routines dgetrf and dgetrs from a C application program. 

    The program generates the matrix, and demonstrates two different ways 
    of accounting for the fact that fortran stores arrays in column major 
    order, while C uses row-major order. To illustrate the differences in 
    the two techniques, a small 3 by 3 matrix is used:

                 18.80      22.00      25.20
                 23.50      26.70      35.00
                 31.80      35.00      40.60

   The right hand side of the problem is generated assuming a
   known solution of all 1.0, and using the matrix-vector multiply
   routine dgemv from CXML. 

   Method 1 takes the matrix above, stored in the row-major scheme,
   and first transposes it. This transposed matrix is used in all
   subsequent calculations.

   Method 2 uses the matrix in the row-major scheme, but instead of
   solving Ax = b, solves (A transpose)*x = b, implicitly taking into
   account the row major vs. column major storage of C vs. Fortran.
   This method avoids the cost of transposing the matrix.

   Note that the LU factors printed out for each method is different
   as in the first case, the matrix is transposed before factorization.
      
   This program illustrates the following:
         - routine naming conventions
         - differences in array indexing between Fortran and C:
                C default x[n]: 0 to (n-1)
                FORTRAN default  x(n): 1 to n
         - handling of two dimensional arrays in Fortran and C.

   Note: the code used in this example works on all Compaq platforms.
   Conditional compilation is used to select the statements appropriate
   to each operating system.

   For more detailed explanation of the routines used, please check 
   the CXML Reference Manual.
*/

/*

       All output is printed to the screen. A sample output follows:

demonstrating lapack calls by transposing the matrix

  problem size: 3

  matrix before transpose
    18.80      22.00      25.20  
    23.50      26.70      35.00  
    31.80      35.00      40.60  

  matrix after transpose
    18.80      23.50      31.80  
    22.00      26.70      35.00  
    25.20      35.00      40.60  

  Right hand side: 
    66.00
    85.20
    107.40

  info after dgetrf 0
  Pivot vector: 
    3
    3
    3

  Matrix after factorization:
    31.80       0.59       0.74  
    35.00       1.31       0.64  
    40.60       1.20       4.23  

  info after dgetrs 0

  Solution vector: 
    1.000000e+00
    1.000000e+00
    1.000000e+00


demonstrating lapack calls without transposing the matrix

  problem size: 3

  original matrix
    18.80      22.00      25.20  
    23.50      26.70      35.00  
    31.80      35.00      40.60  

  Right hand side: 
    66.00
    85.20
    107.40

  info after dgetrf 0
  Pivot vector: 
    3
    2
    3

  Matrix after factorization:
    25.20       0.87       0.75  
    35.00      -3.86       0.68  
    40.60      -0.44       1.81  

  info after dgetrs 0

  Solution vector: 
    1.000000e+00
    1.000000e+00
    1.000000e+00

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
   Add trailing underscores to Fortran routines on linux, unix.
   Add descriptors for VMS
*/

# if (defined( __unix__ ) || (defined(linux)) )

#   define char_arg( a ) a
#   define char_arg_decl char*
#   define __stdcall

#   define dgemv_name dgemv_
#   define dgetrf_name dgetrf_
#   define dgetrs_name dgetrs_

# elif defined( _WIN32 )

#   define char_arg( a ) a, 1
#   define char_arg_decl char*, int

#   define dgemv_name DGEMV
#   define dgetrf_name DGETRF
#   define dgetrs_name DGETRS

# elif defined( vms ) || defined( __vms )

#   define char_arg( a ) a
#   define char_arg_decl char*
#   define __stdcall

#   define dgemv_name dgemv
#   define dgetrf_name dgetrf
#   define dgetrs_name dgetrs

#   include <descrip.h>
    void SetStringDescriptorConst();

# endif

#if defined( __cplusplus )
    extern "C"
    {
#endif

# define dgemv( arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11 ) \
	dgemv_name( char_arg( arg1 ), arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11 )
extern void __stdcall dgemv_name( char_arg_decl, int*, int*, double*, double*, int*, double*, int*, double*, double*, int* ) ;

# define dgetrf( arg1, arg2, arg3, arg4, arg5, arg6 ) \
	dgetrf_name( arg1, arg2, arg3, arg4, arg5, arg6 )
extern void __stdcall dgetrf_name( int*, int*, double*, int*, int*, int* ) ;

# define dgetrs( arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9 ) \
	dgetrs_name( char_arg( arg1 ), arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9 )
extern void __stdcall dgetrs_name( char_arg_decl, int*, int*, double*, int*, int*, double*, int*, int* ) ;

#if defined( __cplusplus )
    }
#endif

#define max_size  10
#define max_nrhs  5

extern void print_matrix();
extern void t_matrix();
extern void init_matrix();

int main()
{
    double *a; 
    double *b;
    double *x;
    double alpha, beta;
    
    int ipvt[max_size]; 
    int length, lda, info; 
    int status; 
    int i, j, cnt, ldb, nrhs; 

#if defined(vms) || defined(__vms)
    struct dsc$descriptor c1;
#endif

/* 
    start of code 
*/

    length = 3; 
    lda = max_size; 
    ldb = max_size; 
    nrhs = 1;

    alpha = 1.0;    
    beta = 0.0;
    
    a = (double *)malloc(max_size*max_size*sizeof(double));
    b = (double *)malloc(max_nrhs*max_size*sizeof(double));
    x = (double *)malloc(max_nrhs*max_size*sizeof(double));

/* 
    method 1: since C stores 2-dimensional arrays in row major 
    order, and fortran expects a column major order, transpose 
    the matrix at the beginning, and perform all operations on 
    the transposed matrix.                                     
*/
    init_matrix(length, a);

    printf("  matrix before transpose\n"); 
    print_matrix(length, a);  

/* 
    transpose the matrix 
*/

    t_matrix(length, a); 

    printf("\n  matrix after transpose\n"); 
    print_matrix(length, a);  

/* 
    obtain the exact solution in order to generate the right hand side 
*/

    for (i=0; i < length; i++)
	{
	    x[0*max_size+i] = 1.0;
	}

/* 
    given the exact solution, calculate the right hand side, using 
    CXML routine dgemv. note that if the first parameter is 'N', 
    dgemv calculates A*x, where A is stored in column-major order.
    as A has been already been transposed, use 'N' for the first  
    parameter.                                                     
*/

#if defined(vms) || defined(__vms)

    SetStringDescriptorConst(&c1,"N");
    dgemv(&c1, &length, &length, &alpha, a, &lda, x, &nrhs,
	   &beta, b, &nrhs);

#else

    dgemv("N", &length, &length, &alpha, a, &lda, x, &nrhs,
	  &beta, b, &nrhs);

#endif
    
    printf("\n  Right hand side: \n");
    for (i = 0; i < length; i++)
	printf ("    %.2f\n",b[0*max_size+i]);

/* 
    call the factorize routine dgetrf 
*/

    dgetrf(&length, &length, a, &lda, ipvt, &info); 
    printf ("\n  info after dgetrf %d\n",info);

    printf("  Pivot vector: \n");
    for (i = 0; i < length; i++)
	printf ("    %d\n",ipvt[i]);
    
    printf("\n  Matrix after factorization:\n"); 
    print_matrix(length, a);  

/* 
    call the solve routine; as matrix was originally transposed 
    this routine is called with first parameter "N".            
*/

#if defined(vms) || defined(__vms)

    SetStringDescriptorConst(&c1,"N");
    dgetrs(&c1, &length, &nrhs, a, &lda, ipvt, b, &ldb, &info);

#else

    dgetrs("No Transpose", &length, &nrhs, a, &lda, ipvt, b, &ldb, &info);

#endif

    printf ("\n  info after dgetrs %d\n",info);

    printf("\n  Solution vector: \n");
    for (i = 0; i < length; i++)
	printf ("    %e\n",b[0*max_size+i]);


/* 
    method 2: use the fact that lapack can solve either Ax=b or 
    (A transpose)*x = b, to solve the system without tranposing 
    the matrix.                                                 
*/

    printf("\n\ndemonstrating lapack calls without transposing the matrix\n");
    printf("\n  problem size: %d\n\n",length);
    
/* 
    initialize the matrix, storing elements in row-major order 
*/

    init_matrix(length, a);

    printf("  original matrix\n"); 
    print_matrix(length, a);  

/* 
    obtain the exact solution in order to generate the right hand side 
*/

    for (i=0; i < length; i++)
	{
	    x[0*max_size+i] = 1.0;
	}

/* 
    given the exact solution, calculate the right hand side, using 
    CXML routine dgemv. note that if the first parameter is 'N',
    dgemv calculates A*x, where A is stored in column-major order.
    as A is in row major order, use 'T' for the first parameter,  
    the routine will calculate (A transpose)*x, thus implicitly   
    taking care of the difference in storage between Fortran and C.
*/


#if defined(vms) || defined(__vms)

    SetStringDescriptorConst(&c1,"T");    
    dgemv(&c1, &length, &length, &alpha, a, &lda, x, &nrhs,
	   &beta, b, &nrhs);

#else

    dgemv("T", &length, &length, &alpha, a, &lda, x, &nrhs,
	   &beta, b, &nrhs);

#endif
    
    printf("\n  Right hand side: \n");
    for (i = 0; i < length; i++)
	printf ("    %.2f\n",b[0*max_size+i]);

/* 
    call the factorize routine dgetrf 
*/

    dgetrf(&length, &length, a, &lda, ipvt, &info); 
    printf ("\n  info after dgetrf %d\n",info);

    printf("  Pivot vector: \n");
    for (i = 0; i < length; i++)
	printf ("    %d\n",ipvt[i]);
    
    printf("\n  Matrix after factorization:\n"); 
    print_matrix(length, a);  

/* 
    call the solve routine; as matrix was not transposed 
    this routine is called with first parameter "T", to  
    take into account the difference in storage between  
    Fortran and C.
*/


#if defined(vms) || defined(__vms)

    SetStringDescriptorConst(&c1,"T");    
    dgetrs(&c1, &length, &nrhs, a, &lda, ipvt, b, &ldb, &info);

#else

    dgetrs("Transpose", &length, &nrhs, a, &lda, ipvt, b, &ldb, &info);

#endif

    printf ("\n  info after dgetrs %d\n",info);

    printf("\n  Solution vector: \n");
    for (i = 0; i < length; i++)
	printf ("    %e\n",b[0*max_size+i]);


    free (a);
    free (b);
    free (x);
    
}  /* end of main() */



/* initialize the matrix */

void init_matrix(int length, double a[])
{
    a[0] = 18.80;
    a[1] = 22.0;
    a[2] = 25.20;
    
    a[max_size+0] = 23.50;
    a[max_size+1] = 26.70;
    a[max_size+2] = 35.00;
    
    a[2*max_size+0] = 31.80;
    a[2*max_size+1] = 35.00;
    a[2*max_size+2] = 40.60;

} /* end of init_matrix */



/* print the matrix */

void print_matrix(int length, double a[])
{
    int i, j; 

    for (i=0; i<length; i++)
    {
        for (j=0; j<length; j++)
            printf("   %6.2f  ", a[i*max_size+j]);
        printf("\n");
    }

}  /* end of print_matrix() */



/* transpose the matrix */        

void t_matrix(int length, double a[])
{
    int i, j; 
    double *tmp;

    /*  Allocate a temporary buffer and save the original data in it.  */
    tmp = (double *)malloc(length*length*sizeof(double));
    for (i=0; i<length; i++)
      for (j=0; j<length; j++)
	tmp[i*length+j] = a[i*max_size+j]; 

    /*  Swap a(i,j) with a(j,i).  */
    for (i=0; i<length; i++)
      for (j=0; j<length; j++)
	a[i*max_size+j] = tmp[j*length+i]; 

    free (tmp);
    
}  /* end of t_matrix() */


/* 
   define the descriptors for VMS, for passing character string arguments 
   to Fortran routines
*/

#if defined(vms) || defined(__vms)

void SetStringDescriptorConst(struct dsc$descriptor *desc,
				char *string)
{
    desc->dsc$a_pointer = string;
    desc->dsc$w_length = strlen(string);
    desc->dsc$b_dtype = DSC$K_DTYPE_T;
    desc->dsc$b_class = DSC$K_CLASS_S;
}

#endif


